      module HierarchicalSolver
      include 'dmumps_struc.h'

c IRN,JCN,A_loc contain local problem (obtained at input, kept)
c BEGIN INTERFACE INPUT
      integer :: N !global N
      integer :: NZ !global NZ
      integer :: NZ_loc !dimension of *_loc
      integer, dimension(:), pointer :: IRN_loc  !global rows on local MPI process
      integer, dimension(:), pointer :: JCN_loc   !global columns on local MPI process
      double precision, dimension(:), pointer :: A_loc
c only host processor
      double precision, dimension(:), pointer :: RHS !RHS on input, solution on output
c END INTERFACE INPUT
      logical, dimension(:), pointer :: INTERFACE_loc2 ! which of IRN_loc are on the interface
      integer, dimension(:), pointer :: IRN_loc2  !local rows on local MPI process
      integer, dimension(:), pointer :: JCN_loc2   !local columns on local MPI process
      double precision, dimension(:), pointer :: A_loc2 
      integer, dimension(:), pointer :: mapl2g !local to global indexing map
      integer :: mapl2g_size
c Proc. 0 for distributing RHS
      integer, dimension(:), pointer :: mapsl2g !local to global indexing maps
      integer :: mapsl2g_size
c temp
      integer, dimension(:), pointer :: maps_sizes
      integer, dimension(:), pointer :: IRN_loc_sorted  !sorted global rows on local MPI process
      integer, dimension(:), pointer :: IRN_loc_sorted_ptr  !pointers to entries from sorted global rows
      integer, dimension(:), pointer :: JCN_loc_sorted  !sorted global columns on local MPI process
      integer, dimension(:), pointer :: JCN_loc_sorted_ptr  !pointers to entries from sorted global columns
      logical, dimension(:), pointer :: INTERFACE_loc_sorted ! which of IRN_loc_sorted are on the interface
c VVVVVVVV
      integer :: COMM_LOC
      integer :: myrank
      integer :: nrproc
      type(dmumps_struc) :: mumps_phaseI
c local MUMPS gets IRN,JCN,A_loc --> return SCHUR complement
c VVVVVVVV
      integer :: NZ_SCHUR_loc !length of global list of non-zero entries for Schur
      integer :: SIZE_SCHUR !local size of the Schur
      integer, dimension(:), pointer :: SCHUR_IRN_loc ! global rows
      integer, dimension(:), pointer :: SCHUR_JCN_loc !global columns      
      integer, dimension(:), pointer :: SCHUR_LISTVAR !Schur variables
      double precision, dimension(:), pointer :: SCHUR_A_loc  
      double precision, dimension(:), pointer :: SCHUR_RHS_loc
      double precision, dimension(:), pointer :: RHS_loc
c temp
      double precision, dimension(:), pointer :: RHS_part
      double precision, dimension(:), pointer :: RHS_tmp
      integer, dimension(:), pointer :: IRN_temp
      integer, dimension(:), pointer :: JCN_temp
c VVVVVVVV
      type(dmumps_struc) :: mumps_global
c global MUMPS gets SCHUR_IRN,JCN,A_loc --> returns solution at RHS
c temp
      double precision, dimension(:), pointer :: SOL_part !used by processor 0 to scather parts of solution
c VVVVVVVV
      double precision, dimension(:), pointer :: SOL_loc

      double precision, parameter :: EPSILON_SCHUR = 1.d-8

      contains
c----------------------------------------------------------------------
      recursive subroutine quicksort_irn_loc(ibeg_in,iend_in)
      implicit none
      integer :: ibeg_in,iend_in
      integer :: ibeg,iend
      integer :: iq
      ibeg=ibeg_in;iend=iend_in
c      write(*,*)'called quicksort ibeg,iend',ibeg,iend
c      read(*,*)
      if((iend-ibeg+1) > 1) then
        call quicksort_irn_partition(ibeg,iend,iq)
c        write(*,*)'called partition ibeg,iend,marker',ibeg,iend,iq
        call quicksort_irn_loc(ibeg,iq-1)
        call quicksort_irn_loc(iq,iend)
      endif
      end subroutine quicksort_irn_loc

      subroutine quicksort_irn_partition(ibeg,iend, marker)
      implicit none
      integer, intent(in) :: ibeg
      integer, intent(in) :: iend
      integer, intent(out) :: marker
      integer :: i, j
      integer :: temp_i
      logical :: temp_l
      integer :: ix      ! pivot point
      ix = IRN_loc_sorted(ibeg-1+1)
      i= 0 
      j= iend-ibeg+1 + 1

      do
        j = j-1
        do
           if (IRN_loc_sorted(ibeg-1+j) <= ix) exit
             j = j-1
        end do
        i = i+1
        do
           if (IRN_loc_sorted(ibeg-1+i) >= ix) exit
           i = i+1
        end do
        if (i < j) then
           ! exchange (ibeg-1+i) and (ibeg-1+j)
           temp_i = IRN_loc_sorted(ibeg-1+i)
           IRN_loc_sorted(ibeg-1+i) = IRN_loc_sorted(ibeg-1+j)
           IRN_loc_sorted(ibeg-1+j) = temp_i
           temp_i = IRN_loc_sorted_ptr(ibeg-1+i)
           IRN_loc_sorted_ptr(ibeg-1+i) = IRN_loc_sorted_ptr(ibeg-1+j)
           IRN_loc_sorted_ptr(ibeg-1+j) = temp_i
c           temp_l = INTERFACE_loc_sorted(ibeg-1+i)
c           INTERFACE_loc_sorted(ibeg-1+i)=INTERFACE_loc_sorted(ibeg-1+j)
c           INTERFACE_loc_sorted(ibeg-1+j) = temp_l
        elseif (i.eq.j) then
           marker = ibeg-1+i+1
           return
        else
           marker = ibeg-1+i
           return
        endif
      end do
      end subroutine quicksort_irn_partition

c----------------------------------------------------------------------
      recursive subroutine quicksort_jcn_loc(ibeg_in,iend_in)
      implicit none
      integer :: ibeg_in,iend_in
      integer :: ibeg,iend
      integer :: i,iq
      ibeg=ibeg_in; iend=iend_in
      if((iend-ibeg+1) > 1) then
        call quicksort_jcn_partition(ibeg,iend,iq)
        call quicksort_jcn_loc(ibeg,iq-1)
        call quicksort_jcn_loc(iq,iend)
      endif
      end subroutine quicksort_jcn_loc

      subroutine quicksort_jcn_partition(ibeg,iend, marker)
      implicit none
      integer, intent(in) :: ibeg
      integer, intent(in) :: iend
      integer, intent(out) :: marker
      integer :: i, j
      integer :: temp_i
      logical :: temp_l
      integer :: ix      ! pivot point
      ix = JCN_loc_sorted(ibeg-1+1)
      i= 0
      j= iend-ibeg+1 + 1

      do
        j = j-1
        do
           if (JCN_loc_sorted(ibeg-1+j) <= ix) exit
             j = j-1
        end do
        i = i+1
        do
           if (JCN_loc_sorted(ibeg-1+i) >= ix) exit
           i = i+1
        end do
        if (i < j) then
           ! exchange (ibeg-1+i) and (ibeg-1+j)
           temp_i = JCN_loc_sorted(ibeg-1+i)
           JCN_loc_sorted(ibeg-1+i) = JCN_loc_sorted(ibeg-1+j)
           JCN_loc_sorted(ibeg-1+j) = temp_i
           temp_i = JCN_loc_sorted_ptr(ibeg-1+i)
           JCN_loc_sorted_ptr(ibeg-1+i) = JCN_loc_sorted_ptr(ibeg-1+j)
           JCN_loc_sorted_ptr(ibeg-1+j) = temp_i
        elseif (i == j) then
           marker = ibeg-1+i+1
           return
        else
           marker = ibeg-1+i
           return
        endif
      end do
      end subroutine quicksort_jcn_partition
c----------------------------------------------------------------------
      subroutine phaseI_1_create_communicators
      implicit none
      include "mpif.h"
      integer :: group_comm_world
      integer :: comm_myrank_local
      integer :: iprint
      integer :: i1,i2,i3
      integer :: ierr
      integer :: group
      integer :: processors(1)

      iprint=0
      if(iprint.eq.1)then
        write(*,*)MYRANK,'MPI_COMM_WORLD',MPI_COMM_WORLD
      endif
      call mpi_comm_group(MPI_COMM_WORLD,group_comm_world,i1)
      call mpi_comm_size(MPI_COMM_WORLD,nrproc,i2)
      call mpi_comm_rank(MPI_COMM_WORLD,myrank,i3)
      if ((i1+i2+i3)/=0) then
        write(*,*)'phaseI_1: error calling mpi'
        stop
      endif
      if(iprint.eq.1)then
        write(*,*)MYRANK,'got group_comm_world',group_comm_world
      endif
      processors(1)=myrank
      if(iprint.eq.1)then
        write(*,*)MYRANK,'extract myrank',myrank
      endif
      call mpi_group_incl(group_comm_world,1,
     .  processors,group,i1)
      if ((i1)/=0) then
        write(*,*)myrank,':phaseI_1:',
     .      'error calling mpi_group_incl'
            stop
      endif
      if(iprint.eq.1)then
        write(*,*)MYRANK,'got group',group
      endif
      call mpi_comm_create(MPI_COMM_WORLD,group,
     .       comm_myrank_local,i2)
      COMM_LOC=comm_myrank_local
      if ((i2)/=0) then
        write(*,*)myrank,':phaseI_1:',
     .        'error calling mpi_com_create'
        stop
      endif
      if(iprint.eq.1)then
        write(*,*)MYRANK,'got COMM_LOC',COMM_LOC
      endif
      end subroutine phaseI_1_create_communicators
c----------------------------------------------------------------------
      subroutine phaseI_2_create_local_mumps
      implicit none
c  ...set an MPI communicator for all MUMPS processors
      mumps_phaseI%comm = COMM_LOC
c
c  ...indicate that host processor should participate in the
c     factorization and solve phases
      mumps_phaseI%par = 1
c
c     symmetric and positive definite (2 = general symmetric)
      mumps_phaseI%sym = 0

      mumps_phaseI%job = -1

      call dmumps(mumps_phaseI)

c     error output stream (non-positive to suppress)
      mumps_phaseI%icntl(1)  = 0 !1
c     diagnostic, statistics and warnings
      mumps_phaseI%icntl(2)  = 0 !1
c     global information
      mumps_phaseI%icntl(3)  = 0 !6
c     printing level
      mumps_phaseI%icntl(4)  = 0 !3
c     input matrix in assembled or element format
      mumps_phaseI%icntl(5)  = 0
c     column permutation for zero-free diagonal (automatic)
      mumps_phaseI%icntl(6)  = 7
c     pivot order (automatic)
      mumps_phaseI%icntl(7)  = 5 !1 enforce ordering, 5 metis, 0 AMD, 7 auto
c     scaling (automatic)
      mumps_phaseI%icntl(8)  = 7
c     no transpose
      mumps_phaseI%icntl(9)  = 1
c     max steps for iterative refinement
      mumps_phaseI%icntl(10) = 0
c     statistics info
      mumps_phaseI%icntl(11) = 0
c     controls parallelism
      mumps_phaseI%icntl(12) = 0
c     use ScaLAPACK for root node
      mumps_phaseI%icntl(13) = 1 !0 use 1 do not use
c     percentage increase in estimated workspace
      mumps_phaseI%icntl(14) = 100 
c     matrix distribution for assembled input
      mumps_phaseI%icntl(18) = 0 !0 local 3 distributed
c     nonzero for Schur complement
      mumps_phaseI%icntl(19) = 1
c     distribution of RHS (centralized on host)
      mumps_phaseI%icntl(20) = 0
c     out-of-core
c      mumps_par_get_schur%icntl(22) = 1
c      mumps_par_get_schur%OOC_TMPDIR = "/workspace/maciek/"
      end subroutine phaseI_2_create_local_mumps
c--------------------------------------------------------------------
      subroutine phaseI_3_translate_global_to_local
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,ifirst,isecond,inumber,irow,id,ifrom
      integer :: istep,ierr,sender,tag,isize
      integer :: recv(1),send(1)
      integer status(MPI_STATUS_SIZE)
      iprint=0
      allocate(IRN_loc_sorted(NZ_loc))
      allocate(IRN_loc_sorted_ptr(NZ_loc))
c      allocate(INTERFACE_loc_sorted(NZ_loc))
      IRN_loc_sorted(1:NZ_loc)=IRN_loc(1:NZ_loc)
c      INTERFACE_loc_sorted(1:NZ_loc)=INTERFACE_loc(1:NZ_loc)
      do i=1,NZ_loc
        IRN_loc_sorted_ptr(i)=i
      enddo
      if(iprint.eq.1)then
        write(*,*)myrank,
     .    'call quicksort_irn_loc',IRN_loc_sorted(1:NZ_loc)
      endif
      call quicksort_irn_loc(1,NZ_loc)
      if(iprint.eq.1)then
        write(*,*)myrank,
     .    'after quicksort_irn_loc',IRN_loc_sorted(1:NZ_loc)
      endif
      allocate(JCN_loc_sorted(NZ_loc))
      allocate(JCN_loc_sorted_ptr(NZ_loc))
      JCN_loc_sorted(1:NZ_loc)=JCN_loc(1:NZ_loc)
      do i=1,NZ_loc
        JCN_loc_sorted_ptr(i)=i
      enddo
      if(iprint.eq.1)then
        write(*,*)myrank,
     .    'call quicksort_jcn_loc',JCN_loc_sorted(1:NZ_loc)
      endif
      call quicksort_jcn_loc(1,NZ_loc)
      if(iprint.eq.1)then
        write(*,*)myrank,
     .    'after quicksort_irn_loc',JCN_loc_sorted(1:NZ_loc)
      endif
c----------------------------------------------------------------
c-> build list of interface flags
c      if(myrank.eq.3)iprint=1
      allocate(INTERFACE_loc2(NZ_loc))
      INTERFACE_loc2=.false.
      allocate(IRN_temp(NZ_loc))
      IRN_temp(1:NZ_loc)=IRN_loc_sorted(1:NZ_loc)
      if(iprint.eq.1)write(*,*)myrank,'exchange IRN_temp',IRN_temp
      isize = NZ_loc
      do istep=1,nrproc-1
         if(iprint.eq.1)then
           write(*,*)myrank,'Loop istep',istep,'isize',isize
         endif
c  .....each processor sends to next one
         if(myrank.lt.nrproc-1)then
           if(iprint==1)then
           write(*,*)myrank,'call mpi_send(',isize,1,
     .      'MPI_INTEGER,',myrank+1,1000000*(myrank+1),')'
           endif
           send(1)=isize
           call mpi_send(send,1,
     .      MPI_INTEGER,myrank+1,1000000*(myrank+1),MPI_COMM_WORLD,ierr)
           if(iprint==1)then
           write(*,*)myrank,'call mpi_send(IRN_temp,',isize,
     .       'MPI_INTEGER,',myrank+1,myrank+1,')'
           endif
           call mpi_send(IRN_temp,isize,
     .       MPI_INTEGER,myrank+1,myrank+1,MPI_COMM_WORLD,ierr)
         else if(myrank.eq.nrproc-1)then
           if(iprint==1)then
           write(*,*)myrank,'call mpi_send(',isize,1,
     .       'MPI_INTEGER',0,100000,')'
           endif           
           send(1)=isize
           call mpi_send(send,1,
     .       MPI_INTEGER,0,100000,MPI_COMM_WORLD,ierr)
           if(iprint==1)then
           write(*,*)myrank,'call mpi_send(IRN_temp,',isize,
     .       'MPI_INTEGER',0,0,')'
           endif
           call mpi_send(IRN_temp,isize,
     .       MPI_INTEGER,0,0,MPI_COMM_WORLD,ierr)
         endif
         deallocate(IRN_temp)
c  ......each processor receives from previous one 
         if(myrank.gt.0)then
           sender=myrank-1; tag=1000000*myrank
           if(iprint==1)then
           write(*,*)myrank,'call mpi_recv(,isize',1,
     .       'MPI_INTEGER',sender,tag,')','istep',istep
           endif
           call mpi_recv(recv,1,
     .       MPI_INTEGER,sender,tag,MPI_COMM_WORLD,status,ierr)
           isize=recv(1)
           if(iprint==1)then
           write(*,*)myrank,'received isize',isize,'istep',istep
           endif
           allocate(IRN_temp(isize))
           sender=myrank-1; tag=myrank
           if(iprint==1)then
           write(*,*)myrank,'call mpi_recv(IRN_temp',isize,
     .       'MPI_INTEGER',sender,tag,')','istep',istep
           endif
           call mpi_recv(IRN_temp,isize,
     .       MPI_INTEGER,sender,tag,MPI_COMM_WORLD,status,ierr)
         else
           sender=nrproc-1; tag=100000
           if(iprint==1)then
           write(*,*)myrank,'call mpi_recv(,isize',1,
     .       'MPI_INTEGER',sender,tag,')','istep',istep
           endif
           call mpi_recv(recv,1,
     .       MPI_INTEGER,sender,tag,MPI_COMM_WORLD,status,ierr)
           isize=recv(1)
           if(iprint==1)then
           write(*,*)myrank,'received isize',isize,'istep',istep
           endif
           allocate(IRN_temp(isize))
           sender=nrproc-1; tag=0
           if(iprint==1)then
           write(*,*)myrank,'call mpi_recv(IRN_temp',isize,
     .       'MPI_INTEGER',sender,tag,')','istep',istep
           endif
           call mpi_recv(IRN_temp,isize,
     .       MPI_INTEGER,sender,tag,MPI_COMM_WORLD,status,ierr)
         endif
c         if(istep.eq.0)stop
         if(iprint.eq.1)then
           write(*,*)myrank,'istep',istep
           write(*,*)myrank,'compare IRN_loc_sorted',IRN_loc_sorted
           write(*,*)myrank,'compare IRN_temp',IRN_temp
           write(*,*)myrank,'before INTERFACE_loc2',INTERFACE_loc2
         endif
c  ......compare both lists (first=IRN_loc_sorted,second=IRN_temp)
c-> MP fix
         do ifirst=1,NZ_loc
             if(IRN_loc_sorted(ifirst).eq.1
     .         .or.IRN_loc_sorted(ifirst).eq.N)then 
c..............turn the flag
               INTERFACE_loc2(ifirst)=.true.
             endif
         enddo
c<- MP fix
         ifirst=1; isecond=1; 
         do i=1,NZ_loc+isize
           if(iprint.eq.1)
     .        write(*,*)myrank,'i',i,'ifirst',ifirst,'isecond',isecond
           if(ifirst.gt.NZ_loc.and.isecond.gt.isize)then
              exit
           else if(ifirst.gt.NZ_loc)then
c............the first vector has ended
             exit
           else if(isecond.gt.NZ_loc)then
c............the second vector has ended
             exit
           else
c............both vectors hasn't ended yet
             if(IRN_loc_sorted(ifirst).eq.IRN_temp(isecond)
     .         )then
c..............turn the flag
               INTERFACE_loc2(ifirst)=.true.
c..............shift both lists
               ifirst=ifirst+1; !isecond=isecond+1
            else if(IRN_loc_sorted(ifirst)
     .       .lt.IRN_temp(isecond))then
c..............shift first list
               ifirst=ifirst+1;
             else if(IRN_loc_sorted(ifirst)
     .       .gt.IRN_temp(isecond))then
c..............shift second list
               isecond=isecond+1;
             else
c..............impossible
               write(*,*)'impossible:IRN_loc_sorted(',ifirst,')',
     .  IRN_loc_sorted(ifirst),'IRN_temp(',isecond,')',IRN_temp(isecond)
               stop
             endif
           endif
        enddo 
         if(iprint.eq.1)then
           write(*,*)myrank,'after INTERFACE_loc2',INTERFACE_loc2
           write(*,*)myrank,'INTERFACE_loc_sorted',INTERFACE_loc_sorted
         endif        
      enddo
      deallocate(IRN_temp)
      iprint=0
      allocate(INTERFACE_loc_sorted(NZ_loc))
      INTERFACE_loc_sorted(1:NZ_loc)=INTERFACE_loc2(1:NZ_loc)
c---------------------------------------------------------------
c<- build list of interface flags  
 10      continue
c......simultanous browse of both lists, count
c......number of contributions (without repetitions)
      if(iprint.eq.1)then
        write(*,*)myrank,'count number of contributions'
      endif
       ifirst=1; isecond=1; inumber=0;
       do i=1,NZ_loc
c-> omit repeated entries         
 1       if(ifirst.lt.NZ_loc)then
           if(IRN_loc_sorted(ifirst+1).eq.IRN_loc_sorted(ifirst))then
             ifirst=ifirst+1
c             if(iprint==1)write(*,*)myrank,'ifirst',ifirst,'go to 1'
             go to 1
           endif
         endif
 2       if(isecond.lt.NZ_loc)then
           if(JCN_loc_sorted(isecond+1).eq.JCN_loc_sorted(isecond))then
             isecond=isecond+1
c             if(iprint==1)write(*,*)myrank,'isecond',isecond,'go to 2'
             go to 2
           endif
         endif
c<- omit repeated entries
         if(iprint==1)then
           if(ifirst.le.NZ_loc)then
             write(*,*)myrank,'first value:',
     .       'IRN_loc_sorted(',ifirst,')=',
     .       IRN_loc_sorted(ifirst)
           endif
           if(isecond.le.NZ_loc)then
             write(*,*)myrank,'second value:',
     .       'JCN_loc_sorted(',isecond,')=',
     .       JCN_loc_sorted(isecond)
           endif
         endif
         if(ifirst.gt.NZ_loc
     .     .and.isecond.gt.NZ_loc)then
            exit
         else if(ifirst.gt.NZ_loc)then
c..........the first vector has ended
c..........just rewrite remaining part of JCN_loc_sorted
c..........count
           inumber=inumber+1;
c..........get location at hypermatrix of the supernode matrix
           if(iprint==1)then
             write(*,*)myrank,
     .         'the first list has ended ->',
     .         'rewrite from the second list, global_id',
     .         JCN_loc_sorted(isecond)
           endif
c..........shift second list
           isecond=isecond+1;
         else if(isecond.gt.NZ_loc)then
c..........the second vector has ended
c..........just rewrite remaining part of IRN_loc_sorted
c..........count
           inumber=inumber+1;
c..........get location at hypermatrix of the supernode matrix
           if(iprint==1)then
             write(*,*)myrank,
     .         'the second list has ended ->',
     .         'rewrite from the first list, global_id',
     .         IRN_loc_sorted(ifirst)
           endif
c..........shift both lists
           ifirst=ifirst+1; isecond=isecond+1;
         else
c..........both vectors hasn't ended yet
         if(IRN_loc_sorted(ifirst)
     .     .eq.JCN_loc_sorted(isecond))then
c..........count
           inumber=inumber+1;
c..........get location at hypermatrix of the supernode matrix
           if(iprint==1)then
             write(*,*)myrank,
     .         'both values are equal ->',
     .         'rewrite from the first list, global_id',
     .         IRN_loc_sorted(ifirst)
           endif
c..........shift both lists
           ifirst=ifirst+1; isecond=isecond+1;
         else if(IRN_loc_sorted(ifirst)
     .     .lt.JCN_loc_sorted(isecond))then
c..........count
           inumber=inumber+1;
c..........get location at hypermatrix of the supernode matrix
           if(iprint==1)then
             write(*,*)myrank,
     .         'first value is lower ->',
     .         'rewrite from the first list, global_id',
     .         IRN_loc_sorted(ifirst)
           endif
c..........shift first list
           ifirst=ifirst+1;
         else if(IRN_loc_sorted(ifirst)
     .     .gt.JCN_loc_sorted(isecond))then
c..........count
           inumber=inumber+1;
c..........get location at hypermatrix of the supernode matrix
           if(iprint==1)then
             write(*,*)myrank,
     .         'first value is greater ->',
     .         'rewrite from the second list, global_id',
     .         JCN_loc_sorted(isecond)
           endif
c..........shift second list
           isecond=isecond+1;
         else
c..........impossible
           stop
         endif
         endif
      enddo
      if(iprint.eq.1)then
        write(*,*)myrank,'found ',inumber
      endif    
c  ...the number of global entries is inumber
      mapl2g_size = inumber
      allocate(mapl2g(inumber))
c  ...make a copy with local ids
      allocate(mumps_phaseI%IRN(NZ_loc))
      allocate(mumps_phaseI%JCN(NZ_loc))
      if(iprint.eq.1)then
        write(*,*)myrank,'fill mapl2g'
      endif
c      if(myrank.eq.0)iprint=1
c.....fill the mapl2g
      ifirst=1; isecond=1; id=0; ifrom=0
      do irow=1,inumber
c-> omit repeated entries
 3       if(ifirst.lt.NZ_loc)then
          if(IRN_loc_sorted(ifirst+1).eq.IRN_loc_sorted(ifirst))then
             mumps_phaseI%IRN(ifirst) = id+1
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%IRN(',ifirst,')=',id+1
             ifirst=ifirst+1
c             if(iprint==1)write(*,*)myrank,'ifirst',ifirst,'go to 3'
             go to 3
           endif
         endif
 4       if(isecond.lt.NZ_loc)then
           if(JCN_loc_sorted(isecond+1).eq.JCN_loc_sorted(isecond))then
             mumps_phaseI%JCN(isecond) = id+1
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%JCN(',isecond,')=',id+1
             isecond=isecond+1
c             if(iprint==1)write(*,*)myrank,'isecond',isecond,'go to 4'
             go to 4
           endif
         endif
c<- omit repeated entries
        if(ifirst.gt.NZ_loc
     .     .and.isecond.gt.NZ_loc)then
            exit
        else if(ifirst.gt.NZ_loc)then
c.........the first list has ended
c.........rewrite from second one
          id=id+1;
          mapl2g(id) = JCN_loc_sorted(isecond)
          mumps_phaseI%JCN(isecond) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%JCN(',isecond,')=',id
c.........shift second list
          isecond=isecond+1;
        else if(isecond.gt.NZ_loc)then
c.........the second list has ended
c.........rewrite from first one
          id=id+1;
          mapl2g(id) = IRN_loc_sorted(ifirst)
          mumps_phaseI%IRN(ifirst) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%IRN(',ifirst,')=',id
c.........shift first list
          ifirst=ifirst+1;
        else
        if(IRN_loc_sorted(ifirst)
     .    .eq.JCN_loc_sorted(isecond))then
c.........rewrite nodes info from first matrix
          id=id+1;
          mapl2g(id) = IRN_loc_sorted(ifirst)
          mumps_phaseI%IRN(ifirst) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%IRN(',ifirst,')=',id
          mumps_phaseI%JCN(isecond) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%JCN(',isecond,')=',id
c.........shift both lists
          ifirst=ifirst+1; isecond=isecond+1
        else if(IRN_loc_sorted(ifirst)
     .    .lt.JCN_loc_sorted(isecond))then
c.........rewrite from first one
          id=id+1;
          mapl2g(id) = IRN_loc_sorted(ifirst)
          mumps_phaseI%IRN(ifirst) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%IRN(',ifirst,')=',id
c.........shift first list
          ifirst=ifirst+1;
        else if(IRN_loc_sorted(ifirst)
     .    .gt.JCN_loc_sorted(isecond))then
c.........rewrite from second one
          id=id+1;
          mapl2g(id) = JCN_loc_sorted(isecond)
          mumps_phaseI%JCN(isecond) = id
          if(iprint==1)write(*,*)myrank,
     .      'mumps_phaseI%JCN(',isecond,')=',id
c.........shift second list
          isecond=isecond+1;
        else
c.........impossible
          stop
        endif
        endif
      enddo
c since originally IRN_loc and JCN_loc was in pairs, we need to "desort"
c      if(myrank.eq.0)then
c         write(*,*)'IRN_loc',IRN_loc
c         write(*,*)'IRN_loc_sorted_ptr',IRN_loc_sorted_ptr
c         write(*,*)'mumps_phaseI%IRN sorted',mumps_phaseI%IRN
c      endif
      allocate(IRN_temp(NZ_loc))
      IRN_temp(1:NZ_loc)=mumps_phaseI%IRN(1:NZ_loc)
      do i =1,NZ_loc
        mumps_phaseI%IRN(IRN_loc_sorted_ptr(i))=IRN_temp(i)
      enddo
c      deallocate(IRN_temp) !we keep it for later, forinitialization of LISTVAR_SCHUR
c      if(myrank.eq.0)then
c         write(*,*)'mumps_phaseI%IRN desorted',mumps_phaseI%IRN
c      endif
      allocate(JCN_temp(NZ_loc))
      JCN_temp(1:NZ_loc)=mumps_phaseI%JCN(1:NZ_loc)
      do i =1,NZ_loc
        mumps_phaseI%JCN(JCN_loc_sorted_ptr(i))=JCN_temp(i)
      enddo
      if(associated(JCN_temp))then
         deallocate(JCN_temp); nullify(JCN_temp)
      endif
      end subroutine phaseI_3_translate_global_to_local
c-------------------------------------------------------------------
      subroutine phaseI_3a_cleanup    
      implicit none
c      if(associated(IRN_loc))then
c         deallocate(IRN_loc); nullify(IRN_loc)
c      endif
c      if(associated(JCN_loc))then
c         deallocate(JCN_loc); nullify(JCN_loc)
c      endif
      end subroutine phaseI_3a_cleanup    
c-------------------------------------------------------------------
      subroutine phaseI_3prim_distribute_rhs
      implicit none
      include "mpif.h"
      integer status(MPI_STATUS_SIZE)
      integer :: iprint
      integer :: i,j,ierr,ibeg
      integer :: sender,tag       
      iprint=0
c  ...gather local maps from other processors
      if(myrank.eq.0)then
        if(iprint==1)write(*,*)myrank,'0:nrproc',nrproc
        allocate(maps_sizes(nrproc))
        maps_sizes(1:nrproc)=0
      else
      endif
      if(iprint==1)
     .  write(*,*)myrank,'call mpi_gather(',mapl2g_size,'1,MPI_INTEGER',
     .                'maps_sizes,1,MPI_INTEGER',
     .                '0,MPI_COMM_WORLD,ierr)'
      call mpi_gather(mapl2g_size,1,MPI_INTEGER,
     .                maps_sizes,1,MPI_INTEGER,
     .                0,MPI_COMM_WORLD,ierr)
      if(myrank.eq.0)then
        mapsl2g_size=0
        do i=1,nrproc
           mapsl2g_size = mapsl2g_size + maps_sizes(i)
        enddo
        if(iprint==1)write(*,*)0,'mapsl2g_size',mapsl2g_size
        allocate(mapsl2g(mapsl2g_size))
      else
      endif
      if(iprint==1)then
      write(*,*)myrank,'mapl2g_size',mapl2g_size
      if(myrank==0)write(*,*)'0,maps_sizes',maps_sizes(1:nrproc)
      endif
      call mpi_barrier(MPI_COMM_WORLD,ierr)
      call mpi_gather(mapl2g,mapl2g_size,MPI_INTEGER,
     .                mapsl2g,mapl2g_size,MPI_INTEGER,
     .                0,MPI_COMM_WORLD,ierr)
      if(iprint==1)then
      if(myrank==0)write(*,*)myrank,'mapsl2g',mapsl2g(1:mapsl2g_size)
      endif
      call mpi_barrier(MPI_COMM_WORLD,ierr)
      if(myrank.eq.0)then
c  .....loop through maps, one by one, and send RHS to the processor
        ibeg=maps_sizes(1)
        do i=2,nrproc
           if(iprint==1)then
             write(*,*)myrank,'allocate(RHS_part(',maps_sizes(i),')'
             write(*,*)myrank,'RHS',RHS
           endif
           allocate(RHS_part(maps_sizes(i)))
           do j=1,maps_sizes(i)
             RHS_part(j)=RHS(mapsl2g(ibeg+j))
           enddo
           ibeg = ibeg + maps_sizes(i)
           if(iprint==1)
     .       write(*,*)myrank,
     .       'call mpi_send(RHS_part(1:',maps_sizes(i),'),',
     .       'MPI_DOUBLE_PRECISION,',i-1,i-1,'MPI_COMM_WORLD,ierr'
           call mpi_send(RHS_part,maps_sizes(i),
     .       MPI_DOUBLE_PRECISION,i-1,i-1,MPI_COMM_WORLD,ierr)
           if(associated(RHS_part))then
              deallocate(RHS_part); nullify(RHS_part)
           endif
        enddo
        allocate(RHS_loc(maps_sizes(1)))
        do j=1,maps_sizes(1)
           RHS_loc(j)=RHS(mapsl2g(j))
         enddo
      else
         allocate(RHS_loc(mapl2g_size))
         if(iprint==1)
     .     write(*,*)myrank,'call mpi_recv(RHS_loc(1:',mapl2g_size,'),',
     .     'MPI_DOUBLE_PRECISION',0,myrank,'MPI_COMM_WORLD,status,ierr)'
         sender=0; tag=myrank
         call mpi_recv(RHS_loc,mapl2g_size,
     .       MPI_DOUBLE_PRECISION,sender,tag,MPI_COMM_WORLD,status,ierr)
      endif
      call mpi_barrier(MPI_COMM_WORLD,ierr)
      end subroutine phaseI_3prim_distribute_rhs
c--------------------------------------------------------------------
      subroutine phaseI_4_submit_Schur_to_mumps
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,ierr,ischur
      iprint=0
      if(iprint==1)then
        write(*,*)myrank,'mumps_phaseI%N = ',mapl2g_size
        write(*,*)myrank,'mumps_phaseI%NZ = ',NZ_loc
        write(*,*)myrank,'mumps_phaseI%NRHS = ',1
      endif
      mumps_phaseI%N = mapl2g_size
      mumps_phaseI%NZ = NZ_loc
      mumps_phaseI%NRHS = 1
      allocate(mumps_phaseI%RHS(mapl2g_size))
      do i=1,mapl2g_size
         mumps_phaseI%RHS(i)=RHS_loc(i)
      enddo
      if(iprint==1)write(*,*)myrank,
     .    'mumps_phaseI%RHS=',RHS_loc(1:mapl2g_size)
      allocate(mumps_phaseI%A(NZ_loc))
      if(iprint==1)write(*,*)myrank,'IRN_loc_sorted=',
     .  IRN_loc_sorted(1:NZ_loc)
      if(iprint==1)write(*,*)myrank,'IRN_loc_sorted_ptr=',
     .  IRN_loc_sorted_ptr(1:NZ_loc)
      if(iprint==1)write(*,*)myrank,'JCN_loc_sorted=',
     .  JCN_loc_sorted(1:NZ_loc)
      if(iprint==1)write(*,*)myrank,'A_loc=',
     .  A_loc(1:NZ_loc)
      do i=1,NZ_loc
         mumps_phaseI%A(i) = A_loc(i)
      enddo
      if(iprint==1)write(*,*)myrank,'phaseI_4:mumps_phaseI%A_loc=',
     .  mumps_phaseI%A(1:NZ_loc)
c ....denote Schur complement
      if(iprint==1)write(*,*)myrank,'INTERFACE_loc_sorted',
     .    INTERFACE_loc_sorted
      call mpi_barrier(MPI_COMM_WORLD,ierr)
      SIZE_SCHUR = 0
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.TRUE.))then
          if(i.lt.NZ_loc)then
             if(IRN_loc_sorted(i).eq.IRN_loc_sorted(i+1))cycle
          endif
          SIZE_SCHUR=SIZE_SCHUR+1
        endif
      enddo
      if(iprint==1)write(*,*)myrank,'SIZE_SCHUR',SIZE_SCHUR
      mumps_phaseI%SIZE_SCHUR = SIZE_SCHUR
      allocate(mumps_phaseI%LISTVAR_SCHUR(SIZE_SCHUR))
      ischur = 0
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.TRUE.))then
          if(i.lt.NZ_loc)then
             if(IRN_loc_sorted(i).eq.IRN_loc_sorted(i+1))cycle
          endif
          ischur=ischur+1
c          write(*,*)myrank,'mumps_phaseI%LISTVAR_SCHUR(',ischur,')=',
c     .      IRN_temp(i)
          mumps_phaseI%LISTVAR_SCHUR(ischur)=IRN_temp(i)
        endif
      enddo
      allocate(mumps_phaseI%SCHUR(SIZE_SCHUR*SIZE_SCHUR))
      end subroutine phaseI_4_submit_Schur_to_mumps
c--------------------------------------------------------------------
      subroutine phaseI_4a_cleanup
      implicit none
      if(associated(IRN_loc_sorted_ptr))then
         deallocate(IRN_loc_sorted_ptr); nullify(IRN_loc_sorted_ptr)
      endif
      if(associated(JCN_loc_sorted_ptr))then
         deallocate(JCN_loc_sorted_ptr); nullify(JCN_loc_sorted_ptr)
      endif
      if(associated(JCN_loc_sorted))then
         deallocate(JCN_loc_sorted); nullify(JCN_loc_sorted)
      endif
      if(associated(IRN_temp))then
         deallocate(IRN_temp); nullify(IRN_temp)
      endif
      end subroutine phaseI_4a_cleanup
c--------------------------------------------------------------------
      subroutine phaseI_5_compute_Schur
      implicit none
c ANALYSIS
      mumps_phaseI%job = 1
      call dmumps(mumps_phaseI)
c FACTORIZATION
      mumps_phaseI%job = 2
      call dmumps(mumps_phaseI)
      end subroutine phaseI_5_compute_Schur
c--------------------------------------------------------------------
      subroutine phaseI_6_get_Schur
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,j,id,index
      iprint=0
c      if(myrank.eq.0)iprint=1
      NZ_SCHUR_loc = 0
      do i=1,mumps_phaseI%SIZE_SCHUR
        do j=1,mumps_phaseI%SIZE_SCHUR
        if(iprint.eq.1)then
        write(*,*)'mumps_phaseI%SCHUR(',i+(j-1)*mumps_phaseI%SIZE_SCHUR,
     .   ')=',mumps_phaseI%SCHUR(i+(j-1)*mumps_phaseI%SIZE_SCHUR)
        endif
          if(abs(mumps_phaseI%SCHUR(i+(j-1)*mumps_phaseI%SIZE_SCHUR))
     .      .gt.EPSILON_SCHUR)then
            NZ_SCHUR_loc = NZ_SCHUR_loc+1
          endif
        enddo 
      enddo
      if(iprint.eq.1)write(*,*)myrank,'NZ_SCHUR_loc',NZ_SCHUR_loc
      if(iprint.eq.1)write(*,*)myrank,'EPSILON_SCHUR',EPSILON_SCHUR
      allocate(SCHUR_IRN_loc(NZ_SCHUR_loc))
      allocate(SCHUR_JCN_loc(NZ_SCHUR_loc))
      allocate(SCHUR_A_loc(NZ_SCHUR_loc))
      id=0
      do i=1,mumps_phaseI%SIZE_SCHUR
        do j=1,mumps_phaseI%SIZE_SCHUR
          if(abs(mumps_phaseI%SCHUR(i+(j-1)*mumps_phaseI%SIZE_SCHUR))
     .      .gt.EPSILON_SCHUR)then
            id=id+1
            SCHUR_IRN_loc(id)=mapl2g(mumps_phaseI%LISTVAR_SCHUR(j))
            SCHUR_JCN_loc(id)=mapl2g(mumps_phaseI%LISTVAR_SCHUR(i))
            SCHUR_A_loc(id)=
     .        mumps_phaseI%SCHUR(i+(j-1)*mumps_phaseI%SIZE_SCHUR)
          endif
        enddo 
      enddo
      allocate(SCHUR_LISTVAR(mumps_phaseI%SIZE_SCHUR))
      SCHUR_LISTVAR(1:mumps_phaseI%SIZE_SCHUR)=
     .  mumps_phaseI%LISTVAR_SCHUR(1:mumps_phaseI%SIZE_SCHUR)
      iprint=0
      if(iprint.eq.1)then
         write(*,*)myrank,'SCHUR_IRN_loc',SCHUR_IRN_loc
         write(*,*)myrank,'SCHUR_JCN_loc',SCHUR_JCN_loc
         write(*,*)myrank,'SCHUR_A_loc',SCHUR_A_loc
      endif
      allocate(SCHUR_RHS_loc(mumps_phaseI%SIZE_SCHUR))
      do i=1,mumps_phaseI%SIZE_SCHUR
         index = mumps_phaseI%LISTVAR_SCHUR(i)
         SCHUR_RHS_loc(i) = mumps_phaseI%RHS(index)
      enddo
      if(iprint.eq.1)write(*,*)myrank,'SCHUR_RHS_loc',SCHUR_RHS_loc
      end subroutine phaseI_6_get_Schur
c----------------------------------------------------------------------
      subroutine phaseI_6prim_copy_local_system
      implicit none
      allocate(IRN_loc2(mumps_phaseI%NZ))
      allocate(JCN_loc2(mumps_phaseI%NZ))
      allocate(A_loc2(mumps_phaseI%NZ))
      IRN_loc2(1:mumps_phaseI%NZ)
     .  = mumps_phaseI%IRN(1:mumps_phaseI%NZ)
      JCN_loc2(1:mumps_phaseI%NZ)
     .  = mumps_phaseI%JCN(1:mumps_phaseI%NZ)
      A_loc2(1:mumps_phaseI%NZ)
     .  = mumps_phaseI%A(1:mumps_phaseI%NZ)
      end subroutine phaseI_6prim_copy_local_system
c----------------------------------------------------------------------
      subroutine phaseI_6a_cleanup_mumps
      implicit none
c  ...terminate an instance of MUMPS
      mumps_phaseI%job = -2
      call dmumps(mumps_phaseI)
      end subroutine phaseI_6a_cleanup_mumps
c----------------------------------------------------------------------
      subroutine phaseI_7_create_global_mumps
      implicit none
      include "mpif.h"
c  ...set an MPI communicator for all MUMPS processors
      mumps_global%comm = MPI_COMM_WORLD
c
c  ...indicate that host processor should participate in the
c     factorization and solve phases
      mumps_global%par = 1
c
c     symmetric and positive definite (2 = general symmetric)
      mumps_global%sym = 0

      mumps_global%job = -1

      call dmumps(mumps_global)

c     error output stream (non-positive to suppress)
      mumps_global%icntl(1)  = 1 !1
c     diagnostic, statistics and warnings
      mumps_global%icntl(2)  = 1 !1
c     global information
      mumps_global%icntl(3)  = 6 !6
c     printing level
      mumps_global%icntl(4)  = 3 !3
c     input matrix in assembled or element format
      mumps_global%icntl(5)  = 0
c     column permutation for zero-free diagonal (automatic)
      mumps_global%icntl(6)  = 7
c     pivot order (automatic)
      mumps_global%icntl(7)  = 5 !1 enforce ordering, 5 metis, 0 AMD, 7 auto
c     scaling (automatic)
      mumps_global%icntl(8)  = 7
c     no transpose
      mumps_global%icntl(9)  = 1
c     max steps for iterative refinement
      mumps_global%icntl(10) = 0
c     statistics info
      mumps_global%icntl(11) = 0
c     controls parallelism
      mumps_global%icntl(12) = 0
c     use ScaLAPACK for root node
      mumps_global%icntl(13) = 0 !0 use 1 do not use
c     percentage increase in estimated workspace
      mumps_global%icntl(14) = 100
c     matrix distribution for assembled input
      mumps_global%icntl(18) = 3 !0 local 3 distributed
c     nonzero for Schur complement
      mumps_global%icntl(19) = 0
c     distribution of RHS (centralized on host)
      mumps_global%icntl(20) = 0
c     out-of-core
c      mumps_par_get_schur%icntl(22) = 1
c      mumps_par_get_schur%OOC_TMPDIR = "/workspace/maciek/"
      end subroutine phaseI_7_create_global_mumps
c--------------------------------------------------------------------
      subroutine phaseI_8_submit_Schur_to_global_mumps      
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,iadd_ones,ierr,ischur
      iadd_ones = mapl2g_size - SIZE_SCHUR
      if(mumps_global%MYID.eq.0)then
        mumps_global%NRHS=1
        mumps_global%N=N
        allocate(mumps_global%RHS(N))
        mumps_global%RHS(1:N)=0.d0
      endif
      allocate(RHS_tmp(N))
      RHS_tmp(1:N)=0.d0
c      write(*,*)myrank,
c     .   'SCHUR_RHS_loc',SCHUR_RHS_loc
      do i=1,SIZE_SCHUR
c         write(*,*)myrank,
c     .      'RHS_tmp(',mapl2g(mapl2g_size-SIZE_SCHUR+i),')=',
c     .      SCHUR_RHS_loc(i)
         RHS_tmp(mapl2g(mapl2g_size-SIZE_SCHUR+i))=SCHUR_RHS_loc(i)
      enddo
c gather the RHS at processor 0
      call mpi_reduce(RHS_tmp,mumps_global%RHS,N,MPI_DOUBLE_PRECISION,
     .                MPI_SUM,0,MPI_COMM_WORLD,ierr)
c      if(myrank.eq.0)write(*,*)myrank,
c     .   'REDUCED mumps_global%RHS',mumps_global%RHS
c      if(myrank.eq.0)write(*,*)myrank,
c     .   'reduced RHS_tmp',RHS_tmp
      if(associated(RHS_tmp))then 
         deallocate(RHS_tmp); nullify(RHS_tmp)
      endif
      allocate(mumps_global%IRN_loc(NZ_SCHUR_loc+iadd_ones))
      allocate(mumps_global%JCN_loc(NZ_SCHUR_loc+iadd_ones))
      allocate(mumps_global%A_loc(NZ_SCHUR_loc+iadd_ones))      
      do i=1,NZ_SCHUR_loc
         mumps_global%IRN_loc(i)=SCHUR_IRN_loc(i) 
         mumps_global%JCN_loc(i)=SCHUR_JCN_loc(i) 
         mumps_global%A_loc(i)=SCHUR_A_loc(i) 
      enddo
      ischur = 0
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.FALSE.))then
          if(i.lt.NZ_loc)then
             if(IRN_loc_sorted(i).eq.IRN_loc_sorted(i+1))cycle
          endif
          ischur=ischur+1
c          write(*,*)'IRN_loc_sorted(',i,')=',IRN_loc_sorted(i)
c          write(*,*)'mumps_global%IRN_loc(',
c     .      NZ_SCHUR_loc+ischur,')=',mapl2g(i)
          mumps_global%IRN_loc(NZ_SCHUR_loc+ischur)=IRN_loc_sorted(i)
          mumps_global%JCN_loc(NZ_SCHUR_loc+ischur)=IRN_loc_sorted(i)
          mumps_global%A_loc(NZ_SCHUR_loc+ischur)=1.d0
        endif
      enddo
      mumps_global%NZ_loc=NZ_SCHUR_loc+iadd_ones
      call mpi_reduce(mumps_global%NZ_loc,NZ,1,MPI_INTEGER,
     .                MPI_SUM,0,MPI_COMM_WORLD,ierr)
      if(myrank.eq.0)then
        mumps_global%NZ=NZ
      endif
      end subroutine phaseI_8_submit_Schur_to_global_mumps      
c--------------------------------------------------------------------
      subroutine phaseI_8a_cleanup      
      end subroutine phaseI_8a_cleanup
c--------------------------------------------------------------------
      subroutine phaseI_9_global_solution
      implicit none
      integer :: iprint
c-> TEST
      iprint=0
      if(iprint==1)then
      write(*,*)myrank,'mumps_global%N',mumps_global%N
      write(*,*)myrank,'mumps_global%NZ',mumps_global%NZ
      write(*,*)myrank,'mumps_global%RHS',mumps_global%RHS
      write(*,*)myrank,'mumps_global%NZ_loc',mumps_global%NZ_loc
      write(*,*)myrank,'mumps_global%IRN_loc',mumps_global%IRN_loc
      write(*,*)myrank,'mumps_global%JCN_loc',mumps_global%JCN_loc
      write(*,*)myrank,'mumps_global%A_loc',mumps_global%A_loc
      endif
c<- TEST
c ANALYSIS
      mumps_global%job = 1
      call dmumps(mumps_global)
c FACTORIZATION
      mumps_global%job = 2
      call dmumps(mumps_global)
c SOLVE
      mumps_global%job = 3
      call dmumps(mumps_global)
      end subroutine phaseI_9_global_solution      
c--------------------------------------------------------------------
      subroutine phaseI_10_scatter_solution
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: sender,tag
      integer :: i,j,ibeg,ierr
      integer status(MPI_STATUS_SIZE)
      iprint=0
c  the result is to have SOL_loc on each processor
      if(mumps_global%MYID.eq.0)then
c  .....loop through maps, one by one, and send the solution to the processor
        ibeg=1
        do i=1,nrproc
           if(iprint.eq.1)write(*,*)myrank,'i=',i
           if(i.eq.1)then
c processor 0 can already create its SOL_loc
             if(iprint.eq.1)write(*,*)myrank,'allocate(SOL_loc(',
     .           maps_sizes(i),')'
             allocate(SOL_loc(maps_sizes(i)))
             do j=1,maps_sizes(i)
               SOL_loc(j)=mumps_global%RHS(mapsl2g(ibeg-1+j))
             enddo
           else
             if(iprint.eq.1)write(*,*)myrank,'allocate(SOL_part(',
     .           maps_sizes(i),')'
             allocate(SOL_part(maps_sizes(i)))
             do j=1,maps_sizes(i)
               SOL_part(j)=mumps_global%RHS(mapsl2g(ibeg-1+j))
             enddo
             ibeg = ibeg + maps_sizes(i)
             if(iprint.eq.1)write(*,*)myrank,'send SOL_part',SOL_part
             if(iprint.eq.1)write(*,*)myrank,'call mpi_send(SOL_part',
     .         maps_sizes(i),
     .         'MPI_DOUBLE_PRECISION',i-1,i-1,'MPI_COMM_WORLD,ierr)'
             call mpi_send(SOL_part,maps_sizes(i),
     .         MPI_DOUBLE_PRECISION,i-1,i-1,MPI_COMM_WORLD,ierr)
             deallocate(SOL_part); nullify(SOL_part)
           endif
        enddo
      else
         if(iprint==1)
     .     write(*,*)myrank,'allocate(SOL_loc(',mapl2g_size,'))'
         allocate(SOL_loc(mapl2g_size))
         if(iprint==1)
     .        write(*,*)myrank,'call mpi_recv(SOL_loc',mapl2g_size,
     .       'MPI_DOUBLE_PRECISION,0,',myrank,'MPI_COMM_WORLD,ierr)'
         sender=0;tag=myrank
         call mpi_recv(SOL_loc,mapl2g_size,
     .      MPI_DOUBLE_PRECISION,sender,tag,MPI_COMM_WORLD,status,ierr)
         if(iprint==1)write(*,*)myrank,'received',SOL_loc
      endif
      call mpi_barrier(MPI_COMM_WORLD,ierr)
      end subroutine phaseI_10_scatter_solution
c--------------------------------------------------------------------
      subroutine phaseI_10a_cleanup_global_mumps
      implicit none
c  ...terminate an instance of MUMPS
      mumps_global%job = -2
      call dmumps(mumps_global)
c ...TO DO clean the rest
      end subroutine phaseI_10a_cleanup_global_mumps
c--------------------------------------------------------------------
      subroutine phaseI_11_initialize_mumps_for_local_solve
      implicit none
c  ...set an MPI communicator for all MUMPS processors
      mumps_phaseI%comm = COMM_LOC
c
c  ...indicate that host processor should participate in the
c     factorization and solve phases
      mumps_phaseI%par = 1
c
c     symmetric and positive definite (2 = general symmetric)
      mumps_phaseI%sym = 0

      mumps_phaseI%job = -1

      call dmumps(mumps_phaseI)

c     error output stream (non-positive to suppress)
      mumps_phaseI%icntl(1)  = 0 !1
c     diagnostic, statistics and warnings
      mumps_phaseI%icntl(2)  = 0 !1
c     global information
      mumps_phaseI%icntl(3)  = 0 !6
c     printing level
      mumps_phaseI%icntl(4)  = 0 !3
c     input matrix in assembled or element format
      mumps_phaseI%icntl(5)  = 0
c     column permutation for zero-free diagonal (automatic)
      mumps_phaseI%icntl(6)  = 7
c     pivot order (automatic)
      mumps_phaseI%icntl(7)  = 5 !1 enforce ordering, 5 metis, 0 AMD, 7 auto
c     scaling (automatic)
      mumps_phaseI%icntl(8)  = 7
c     no transpose
      mumps_phaseI%icntl(9)  = 1
c     max steps for iterative refinement
      mumps_phaseI%icntl(10) = 0
c     statistics info
      mumps_phaseI%icntl(11) = 0
c     controls parallelism
      mumps_phaseI%icntl(12) = 0
c     use ScaLAPACK for root node
      mumps_phaseI%icntl(13) = 1 !0 use 1 do not use
c     percentage increase in estimated workspace
      mumps_phaseI%icntl(14) = 100
c     matrix distribution for assembled input
      mumps_phaseI%icntl(18) = 0 !0 local 3 distributed
c     nonzero for Schur complement
      mumps_phaseI%icntl(19) = 0
c     distribution of RHS (centralized on host)
      mumps_phaseI%icntl(20) = 0
c     out-of-core
c      mumps_par_get_schur%icntl(22) = 1
c      mumps_par_get_schur%OOC_TMPDIR = "/workspace/maciek/"

      end subroutine phaseI_11_initialize_mumps_for_local_solve
c--------------------------------------------------------------------
      subroutine phaseI_12_submit_problem_to_mumps
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,id,ierr,NZ_loc_1_schur
      iprint=0
      mumps_phaseI%N = mapl2g_size
      NZ_loc_1_schur=0
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.FALSE.))then            
          NZ_loc_1_schur = NZ_loc_1_schur + 1
        endif
      enddo
      if(iprint.eq.1)write(*,*)myrank,'mumps_phaseI%NZ=',NZ_loc_1_schur
      NZ_loc_1_schur=NZ_loc_1_schur+SIZE_SCHUR
      if(iprint.eq.1)write(*,*)myrank,'mumps_phaseI%NZ+SIZE_SCHUR=',
     .   NZ_loc_1_schur
      mumps_phaseI%NZ = NZ_loc_1_schur
      mumps_phaseI%NRHS = 1
      allocate(mumps_phaseI%RHS(mapl2g_size))
      id=0
      do i=1,NZ_loc
        if(i.lt.NZ_loc)then
          if(IRN_loc_sorted(i).eq.IRN_loc_sorted(i+1))cycle
        endif
        id=id+1
        if(INTERFACE_loc_sorted(i).eqv.(.TRUE.))then
          mumps_phaseI%RHS(id)=SOL_loc(IRN_loc_sorted(i))
        else
          mumps_phaseI%RHS(id)=RHS_loc(IRN_loc_sorted(i))
        endif
      enddo
      do i=1,mapl2g_size         
         mumps_phaseI%RHS(i)=SOL_loc(i)
      enddo
      if(iprint.eq.1)
     .   write(*,*)myrank,'mumps_phaseI%RHS',mumps_phaseI%RHS
c IRN_loc,JCN_loc were associated and filled in step 3
      if(iprint.eq.1)
     .   write(*,*)myrank,'allocate(mumps_phaseI%A,IRN,JCN(',
     .     NZ_loc_1_schur,'))'
      allocate(mumps_phaseI%A(NZ_loc_1_schur))
      allocate(mumps_phaseI%IRN(NZ_loc_1_schur))
      allocate(mumps_phaseI%JCN(NZ_loc_1_schur))
      id=0
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.FALSE.))then
          id=id+1
c          write(*,*)myrank,'id',id
          mumps_phaseI%IRN(id) = IRN_loc2(i)
          mumps_phaseI%JCN(id) = JCN_loc2(i)
          mumps_phaseI%A(id) = A_loc2(i)
        endif
      enddo
c      write(*,*)myrank,'add ones,SIZE SCHUR',SIZE_SCHUR 
      do i=1,NZ_loc
        if(INTERFACE_loc_sorted(i).eqv.(.TRUE.))then
          if(i.lt.NZ_loc)then
             if(IRN_loc_sorted(i).eq.IRN_loc_sorted(i+1))cycle
          endif
          id=id+1
c          write(*,*)myrank,'id',id
          mumps_phaseI%IRN(id) = IRN_loc2(i)
          mumps_phaseI%JCN(id) = IRN_loc2(i)
          mumps_phaseI%A(id) = 1.d0
        endif
      enddo 
      call mpi_barrier(MPI_COMM_WORLD,ierr)      
      end subroutine phaseI_12_submit_problem_to_mumps
c--------------------------------------------------------------------
      subroutine phaseI_12a_cleanup
      implicit none
      if(associated(IRN_loc_sorted))then 
        deallocate(IRN_loc_sorted); nullify(IRN_loc_sorted) 
      endif
      if(associated(INTERFACE_loc_sorted))then
        deallocate(INTERFACE_loc_sorted); nullify(INTERFACE_loc_sorted) 
      endif
      if(associated(SOL_loc))then 
        deallocate(SOL_loc); nullify(SOL_loc)
      endif
      if(associated(RHS_loc))then 
        deallocate(RHS_loc); nullify(RHS_loc)
      endif
      if(associated(IRN_loc2))then 
        deallocate(IRN_loc2); nullify(IRN_loc2)
      endif
      if(associated(JCN_loc2))then 
        deallocate(JCN_loc2); nullify(JCN_loc2)
      endif
      if(associated(A_loc2))then 
        deallocate(A_loc2); nullify(A_loc2)
      endif
      end subroutine phaseI_12a_cleanup
c--------------------------------------------------------------------
      subroutine phaseI_13_compute
      implicit none
c ANALYSIS
      mumps_phaseI%job = 1
      call dmumps(mumps_phaseI)
c FACTORIZATION
      mumps_phaseI%job = 2
      call dmumps(mumps_phaseI)
c SOLVE
      mumps_phaseI%job = 3
      call dmumps(mumps_phaseI)
      end subroutine phaseI_13_compute
c--------------------------------------------------------------------
      subroutine phaseI_14_return_solution
      implicit none
      include "mpif.h"
      integer :: iprint
      integer :: i,j,ibeg,ierr,impi,sender,tag
      integer status(MPI_STATUS_SIZE)
      iprint=0
c solution is collected at RHS in processor 0  
      if(mumps_global%MYID.eq.0)then
c  .....loop through maps, one by one, and send the receive the solution from processor
        ibeg=1
        do i=2,nrproc
           do j=1,maps_sizes(i)
             RHS(mapsl2g(ibeg-1+j))=mumps_phaseI%RHS(j)
           enddo           
           allocate(SOL_part(maps_sizes(i)))
           if(iprint.eq.1)then
           write(*,*)myrank,'call mpi_recv(SOL_part,',maps_sizes(i),
     .    'MPI_DOUBLE_PRECISION,',i-1,i-1,'MPI_COMM_WORLD,status,ierr)'
           endif
           sender=i-1
           tag=i-1
           call mpi_recv(SOL_part,maps_sizes(i),
     .       MPI_DOUBLE_PRECISION,sender,tag,MPI_COMM_WORLD,status,ierr)
           if(iprint.eq.1)write(*,*)myrank,'received',SOL_part
           do j=1,maps_sizes(i)
             RHS(mapsl2g(ibeg-1+j))=SOL_part(j)
           enddo
           if(iprint.eq.1)write(*,*)myrank,'RHS upadted:',RHS
           ibeg = ibeg + maps_sizes(i)
           deallocate(SOL_part); nullify(SOL_part)
        enddo
      else
         if(iprint.eq.1)then
         write(*,*)myrank,'call mpi_send(mumps_phaseI%RHS',mapl2g_size,
     .       'MPI_DOUBLE_PRECISION,0,',myrank,'MPI_COMM_WORLD,ierr)'
         endif
         call mpi_send(mumps_phaseI%RHS(1:mapl2g_size),mapl2g_size,
     .       MPI_DOUBLE_PRECISION,0,myrank,MPI_COMM_WORLD,ierr)
      endif
      end subroutine phaseI_14_return_solution
c--------------------------------------------------------------------
      subroutine phaseI_14a_shutdown
      implicit none
c  ...terminate an instance of MUMPS
      mumps_phaseI%job = -2
      call dmumps(mumps_phaseI)
c  ...cleanup
      if(associated(mapl2g))deallocate(mapl2g)
      if(associated(mapsl2g))deallocate(mapsl2g)
      if(associated(maps_sizes))deallocate(maps_sizes)
      if(associated(IRN_loc_sorted))deallocate(IRN_loc_sorted)
      if(associated(IRN_loc_sorted_ptr))deallocate(IRN_loc_sorted_ptr)
      if(associated(JCN_loc_sorted))deallocate(JCN_loc_sorted)
      if(associated(JCN_loc_sorted_ptr))deallocate(JCN_loc_sorted_ptr)
      if(associated(INTERFACE_loc_sorted))
     .    deallocate(INTERFACE_loc_sorted)
      if(associated(INTERFACE_loc2))
     .    deallocate(INTERFACE_loc2)
      if(associated(SCHUR_IRN_loc))deallocate(SCHUR_IRN_loc)
      if(associated(SCHUR_JCN_loc))deallocate(SCHUR_JCN_loc)
      if(associated(SCHUR_A_loc))deallocate(SCHUR_A_loc)
      if(associated(SOL_part))deallocate(SOL_part)
      end subroutine phaseI_14a_shutdown
c--------------------------------------------------------------------
      end module HierarchicalSolver
